use std::time::Instant;

struct SField {
    field: Vec<usize>,
    rows: Vec<Vec<usize>>,
    cols: Vec<Vec<usize>>,
    blocks: Vec<Vec<usize>>,
    fixed: Vec<usize>,
    size: usize,
    block_size: usize,
    pos: usize,
    pos_last_placed: usize,
    num_fields: usize,
    possible_values: Vec<Vec<usize>>,
}

impl SField {
    pub fn new(block_size: usize) -> SField {
        let size = block_size * block_size;

        // EMPTY FIELD
        // let field = vec![0; size * size];

        // BLOCK_SIZE 4
        // #[rustfmt::skip]
        // let field = vec![
        //     0,15,0,0,0,0,11,4,0,5,0,0,0,0,12,8,
        //     12,0,0,9,0,1,0,5,0,0,8,15,0,0,0,13,
        //     0,0,3,0,0,10,13,0,0,11,4,0,15,0,6,0,
        //     13,10,0,11,0,0,0,14,0,3,2,0,0,9,0,0,
        //     0,0,15,10,8,0,0,0,0,0,14,0,0,6,2,0,
        //     0,0,14,0,0,0,6,0,0,0,0,0,1,0,0,0,
        //     0,11,0,8,3,0,15,1,6,0,0,0,0,0,0,7,
        //     0,6,0,7,0,0,0,0,8,13,0,0,10,0,0,0,
        //     0,14,0,2,0,0,0,0,3,0,5,0,11,15,9,0,
        //     0,0,0,12,11,0,2,0,0,0,0,0,0,0,0,0,
        //     0,8,0,0,6,14,1,0,13,15,0,0,0,0,0,10,
        //     10,0,0,3,9,0,7,0,0,1,0,0,13,12,8,0,
        //     7,0,0,0,0,0,14,0,0,0,0,0,0,0,0,9,
        //     0,3,0,14,0,0,9,6,0,0,0,0,0,4,0,0,
        //     0,4,6,0,0,7,0,0,0,0,0,0,0,0,0,0,
        //     5,2,0,0,0,4,10,15,1,0,3,0,0,0,7,0,
        // ];

        // MANY SOLUTIONS
        #[rustfmt::skip]
        let field = vec![
            0,0,0,0,0,0,0,0,9,
            0,0,0,0,8,9,0,2,0,
            0,0,0,0,2,0,4,0,0,
            0,0,4,0,6,0,0,0,8,
            0,0,0,5,0,0,0,0,0,
            0,6,5,0,0,2,0,7,4,
            0,3,0,0,0,5,0,4,0,
            0,0,1,8,0,0,0,0,0,
            0,0,8,2,0,0,0,6,0,
        ];

        // HARD
        // #[rustfmt::skip]
        // let field = vec![
        //     8,4,0,0,6,0,5,0,1,
        //     0,0,0,0,0,3,0,4,0,
        //     0,0,6,9,0,0,0,0,7,
        //     0,2,0,7,1,0,0,0,6,
        //     0,0,0,6,3,0,0,0,0,
        //     9,0,0,0,0,0,0,5,0,
        //     0,0,0,0,4,0,0,6,0,
        //     2,0,0,0,0,0,1,8,0,
        //     0,0,5,0,0,0,3,0,0,
        // ];

        // EASY
        // #[rustfmt::skip]
        // let field = vec![
        //     5,9,0,6,1,3,0,0,0,
        //     0,0,0,9,0,0,5,0,0,
        //     8,0,3,0,5,7,6,4,0,
        //     0,7,5,0,0,0,4,0,6,
        //     0,6,0,7,4,0,2,0,8,
        //     2,0,8,0,0,0,7,5,3,
        //     0,0,0,5,6,1,0,7,0,
        //     0,0,1,0,7,0,9,0,0,
        //     7,3,6,0,2,0,0,0,0,
        // ];

        let mut fixed = Vec::with_capacity(size * size);
        field.iter().for_each(|e| {
            if *e == 0 {
                fixed.push(0);
            } else {
                fixed.push(1);
            }
        });

        let rows = SField::rows_from_plain_field(&field);
        let cols = SField::cols_from_plain_field(&field);
        let blocks = SField::blocks_from_plain_field(&field);

        SField {
            field,
            rows,
            cols,
            blocks,
            fixed,
            size,
            block_size,
            pos: 0,
            pos_last_placed: 0,
            num_fields: size * size,
            possible_values: vec![vec![]; size * size],
        }
    }

    fn rows_from_plain_field(field: &[usize]) -> Vec<Vec<usize>> {
        let mut rows = vec![vec![0; 9]; 9];
        for row in 0..9 {
            for col in 0..9 {
                rows[row][col] = field[row * 9 + col];
            }
        }
        rows
    }

    fn cols_from_plain_field(field: &[usize]) -> Vec<Vec<usize>> {
        let mut cols = vec![vec![0; 9]; 9];
        for row in 0..9 {
            for col in 0..9 {
                cols[col][row] = field[row * 9 + col];
            }
        }
        cols
    }

    fn blocks_from_plain_field(field: &[usize]) -> Vec<Vec<usize>> {
        fn get_block_with_pos(field: &[usize], pos: usize) -> Vec<usize> {
            let block_start_row = pos / 9 / 3 * 3;
            let block_start_col = pos % 9 / 3 * 3;

            let mut block = vec![];

            for r in 0..3 {
                for c in 0..3 {
                    block.push(field[9 * (block_start_row + r) + block_start_col + c])
                }
            }

            block
        }

        let mut blocks = vec![];
        let block_start_positions = [0, 3, 6, 27, 30, 33, 54, 57, 60];
        for pos in block_start_positions {
            blocks.push(get_block_with_pos(field, pos));
        }

        blocks
    }

    fn build_possible_values_db(&mut self) {
        for idx in 0..self.num_fields {
            if self.fixed[idx] == 1 {
                continue;
            }
            self.pos = idx;
            // try all values between 1 and =self.size and remember the good ones
            let mut good_ones = Vec::with_capacity(self.size);
            for nr in 1..=self.size {
                if self.ok(nr) {
                    good_ones.push(nr);
                }
            }
            self.possible_values[idx] = good_ones;
        }
        self.pos = 0;
    }

    pub fn solve_backtracking(&mut self) -> bool {
        let mut found_solution = false;
        let mut num_solutions = 0;

        //let mut loop_count = 0;
        loop {
            //std::thread::sleep_ms(30);
            //self.print_clear();
            //self.print();
            //loop_count += 1;
            //if loop_count % 1000000 == 0 {
            //self.print_clear();
            //self.print();
            //}
            if self.is_end() {
                found_solution = true;
                num_solutions += 1;
                if num_solutions % 100 == 0 {
                    self.print_clear();
                    println!("Solutions: {}", num_solutions);
                    self.print();
                }

                if self.is_fixed() {
                    continue;
                }
                self.clear_current_field();
                self.prev();
                self.goto_prev_free_field();
            }

            if self.is_fixed() {
                self.next();
                continue;
            }

            if self.goto_next_free_field() {
                if self.put_valid_nr() {
                    self.next();
                    continue;
                } else {
                    //println!("put_valid_nr failed for pos {}", self.pos);
                    //std::thread::sleep_ms(300);
                    self.clear_current_field();
                    if !self.prev() {
                        println!("Number of solutions: {}", num_solutions);
                        break;
                    }
                    if !self.goto_prev_free_field() {
                        println!("Number of solutions: {}", num_solutions);
                        break;
                    }
                }
            }
        }

        found_solution
    }

    #[inline(always)]
    fn print_gray(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x30\x6d");
    }

    #[inline(always)]
    fn print_green(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x32\x6d");
    }

    #[inline(always)]
    fn print_red(&self) {
        print!("\x1b\x5b\x31\x3b\x33\x31\x6d");
    }

    #[inline(always)]
    fn print_neutral(&self) {
        print!("\x1b\x5b\x31\x3b\x30\x6d");
    }

    #[inline(always)]
    fn print_clear(&self) {
        print!("\x1b\x5b\x48\x1b\x5b\x32\x4a");
    }

    #[inline(always)]
    fn print(&self) {
        for i in 0..self.num_fields {
            if i != 0 && i % self.size == 0 {
                println!();
            }

            if i == self.pos_last_placed {
                self.print_red();
            } else if i == self.pos {
                self.print_green();
            } else if self.get_field_at_pos(i) == 0 {
                self.print_gray();
            }

            print!("{:2} ", self.get_field_at_pos(i));

            if i == self.pos || i == self.pos_last_placed || self.get_field_at_pos(i) == 0 {
                self.print_neutral();
            }
        }
        println!();
    }

    #[inline(always)]
    fn clear_current_field(&mut self) {
        self.set(0);
    }

    #[inline(always)]
    fn goto_prev_free_field(&mut self) -> bool {
        while {
            if !self.is_fixed() {
                return true;
            }
            self.prev()
        } {}

        false
    }

    #[inline(always)]
    fn goto_next_free_field(&mut self) -> bool {
        while {
            if !self.is_fixed() {
                return true;
            }
            self.next()
        } {}

        false
    }

    #[inline(always)]
    fn _put_valid_nr(&mut self) -> bool {
        let current_nr = self.get_field_at_pos(self.pos);

        for nr in current_nr..self.size + 1 {
            if self.ok(nr) {
                self.set(nr);
                return true;
            }
        }

        false
    }

    #[inline(always)]
    fn put_valid_nr(&mut self) -> bool {
        let current_nr = self.get_field_at_pos(self.pos);

        // safety:
        //     self.possible_vals is initialized with self.size * self.size
        //     so self.pos can safely be used to index here
        let possible_vals = unsafe { self.possible_values.get_unchecked(self.pos) };

        for nr in possible_vals.iter() {
            if *nr <= current_nr {
                continue;
            }

            if self.ok(*nr) {
                self.set(*nr);
                return true;
            }
        }
        false
    }

    #[inline(always)]
    fn is_end(&self) -> bool {
        !self.field.contains(&0)
    }

    #[inline(always)]
    fn ok(&self, nr: usize) -> bool {
        self.block_ok(nr) && self.row_ok(nr) && self.col_ok(nr)
    }

    #[inline(always)]
    fn _pos_to_xy(&self) -> (usize, usize) {
        let y = self.pos / self.size;
        let x = self.pos % self.size;
        (x + 1, y + 1)
    }

    #[inline(always)]
    fn is_fixed(&self) -> bool {
        // safety: self.pos can be used to index the field unchecked
        //         since the only methods modifying self.pos are
        //         `next()` and `prev()` and they do bounds checking
        unsafe { *self.fixed.get_unchecked(self.pos) == 1 }
    }

    #[inline(always)]
    fn get_field_at_pos(&self, pos: usize) -> usize {
        // safety:
        //     TODO
        //     would need to mathematically explain that the calculations in
        //     get_row(), get_col() and get_block() can never exceed 0..self.num_fields
        unsafe { *self.field.get_unchecked(pos) }
    }

    #[inline(always)]
    fn get_x_y_block_blockoffset(&self) -> (usize, usize, usize, usize) {
        // TODO:
        // this only works for 3x3 sudoku
        #[rustfmt::skip]
        let block_nrs = [
             0, 1, 2, 9,10,11,18,19,20,
             3, 4, 5,12,13,14,21,22,23,
             6, 7, 8,15,16,17,24,25,26,
            27,28,29,36,37,38,45,46,47,
            30,31,32,39,40,41,48,49,50,
            33,34,35,42,43,44,51,52,53,
            54,55,56,63,64,65,72,73,74,
            57,58,59,66,67,68,75,76,77,
            60,61,62,69,70,71,78,79,80,
        ];

        let y = self.pos / self.size;
        let x = self.pos % self.size;

        let block_field = unsafe { block_nrs.get_unchecked(self.pos) };

        let block_nr = block_field / self.size;
        let block_offset = block_field - (block_nr * self.size);

        (x, y, block_nr, block_offset)
    }

    #[inline(always)]
    fn set(&mut self, nr: usize) {
        //assert_eq!(self.field, self.flatten_rows());
        //assert_eq!(self.field, self.flatten_cols());
        //assert_eq!(self.field, self.flatten_blocks());

        self.field[self.pos] = nr;
        self.pos_last_placed = self.pos;

        // update rows, cols and blocks
        // this is more expensive than only updating field but should be outweight
        // by the fact that row_ok(), col_ok() and block_ok() just need to do simple
        // lookups now

        let (x, y, block_nr, block_offset) = self.get_x_y_block_blockoffset();

        //println!("[{},{}] setting {}", y, x, nr);

        let row = unsafe { self.rows.get_unchecked_mut(y) };
        let value = unsafe { row.get_unchecked_mut(x) };
        *value = nr;

        let col = unsafe { self.cols.get_unchecked_mut(x) };
        let value = unsafe { col.get_unchecked_mut(y) };
        *value = nr;

        let block = unsafe { self.blocks.get_unchecked_mut(block_nr) };
        let value = unsafe { block.get_unchecked_mut(block_offset) };
        *value = nr;

        //assert_eq!(self.field, self.flatten_rows());
        //assert_eq!(self.field, self.flatten_cols());
        //assert_eq!(self.field, self.flatten_blocks());
    }

    #[inline(always)]
    fn block_ok(&self, nr: usize) -> bool {
        // TODO
        // works only on 3x3 sudoku
        #[rustfmt::skip]
        let block_nrs = [
            0,0,0,1,1,1,2,2,2,
            0,0,0,1,1,1,2,2,2,
            0,0,0,1,1,1,2,2,2,
            3,3,3,4,4,4,5,5,5,
            3,3,3,4,4,4,5,5,5,
            3,3,3,4,4,4,5,5,5,
            6,6,6,7,7,7,8,8,8,
            6,6,6,7,7,7,8,8,8,
            6,6,6,7,7,7,8,8,8,
        ];

        let block_nr = unsafe { block_nrs.get_unchecked(self.pos) };
        //println!("Block {}: {:?}", block_nr, self.blocks[*block_nr]);
        !unsafe { self.blocks.get_unchecked(*block_nr) }.contains(&nr)
    }

    #[inline(always)]
    fn row_ok(&self, nr: usize) -> bool {
        let row_nr = self.pos / self.size;
        //println!("{:?}", self.rows[row_nr]);
        !unsafe { self.rows.get_unchecked(row_nr) }.contains(&nr)
    }

    #[inline(always)]
    fn col_ok(&self, nr: usize) -> bool {
        let col_nr = self.pos % self.size;
        //println!("{:?}", self.cols[col_nr]);
        !unsafe { self.cols.get_unchecked(col_nr) }.contains(&nr)
    }

    #[inline(always)]
    fn next(&mut self) -> bool {
        if self.pos == self.num_fields - 1 {
            return false;
        }

        self.pos += 1;
        true
    }

    #[inline(always)]
    fn prev(&mut self) -> bool {
        if self.pos == 0 {
            return false;
        }

        self.pos -= 1;
        true
    }

    #[inline(always)]
    fn solve(&mut self) {
        let now = Instant::now();
        if !self.solve_backtracking() {
            println!("there is no solution.");
        }
        println!("took {:.3} seconds", now.elapsed().as_secs_f64());
    }

    //fn flatten_blocks(&self) -> Vec<usize> {
    //#[rustfmt::skip]
    //let block_nrs = [
    //0, 1, 2, 9,10,11,18,19,20,
    //3, 4, 5,12,13,14,21,22,23,
    //6, 7, 8,15,16,17,24,25,26,
    //27,28,29,36,37,38,45,46,47,
    //30,31,32,39,40,41,48,49,50,
    //33,34,35,42,43,44,51,52,53,
    //54,55,56,63,64,65,72,73,74,
    //57,58,59,66,67,68,75,76,77,
    //60,61,62,69,70,71,78,79,80,
    //];

    //let mut flattened = Vec::with_capacity(self.num_fields);
    //for (idx, nr) in block_nrs.iter().enumerate() {
    //let block_nr = nr / self.size;
    //let block_offset = nr - (block_nr * self.size);
    //flattened.push(self.blocks[block_nr][block_offset]);
    //}
    //flattened
    //}

    //fn flatten_rows(&self) -> Vec<usize> {
    //let mut flattened = Vec::with_capacity(self.num_fields);
    //for row in self.rows.iter() {
    //for val in row {
    //flattened.push(*val);
    //}
    //}
    //flattened
    //}

    //fn flatten_cols(&self) -> Vec<usize> {
    //let mut flattened = Vec::with_capacity(self.num_fields);
    //for idx in 0..self.num_fields {
    //flattened.push(self.cols[idx % self.size][idx / self.size]);
    //}
    //flattened
    //}
}

fn run() -> Result<(), String> {
    let mut field = SField::new(3);

    field.build_possible_values_db();

    field.solve();
    Ok(())
}

fn main() {
    if let Err(e) = run() {
        println!("{}", e)
    }
}
